@@ -9,6 +9,7 @@
 #include <unistd.h>
 #include <stdio.h>
 #include <glib.h>
+#include <glib/gprintf.h>
 #include <errno.h>
 #include <dirent.h>
 #include <libxml/parser.h>
@@ -31,6 +32,14 @@
 
 #define MIME_ERROR g_quark_from_static_string("mime-error-quark")
 
+#ifndef PATH_SEPARATOR
+# ifdef _WIN32
+#  define PATH_SEPARATOR ";"
+# else
+#  define PATH_SEPARATOR ":"
+# endif
+#endif
+
 /* This is the list of directories to scan when finding old type files to
  * delete. It is also used to warn about invalid MIME types.
  */
@@ -44,6 +53,8 @@ const char *media_types[] = {
 	"message",
 	"model",
 	"multipart",
+	"x-content",
+	"x-epoc",
 };
 
 /* Represents a MIME type */
@@ -100,13 +111,26 @@ static GHashTable *subclass_hash = NULL;
 /* Maps aliases to Types */
 static GHashTable *alias_hash = NULL;
 
+/* Lists enabled log levels */
+static GLogLevelFlags enabled_log_levels = G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING;
+
 /* Static prototypes */
 static Magic *magic_new(xmlNode *node, Type *type, GError **error);
 
+static void g_log_handler (const gchar   *log_domain,
+			   GLogLevelFlags log_level,
+			   const gchar   *message,
+			   gpointer       unused_data)
+{
+    if (log_level & enabled_log_levels) {
+        g_printf("%s\n", message);
+    }
+}
+
 
 static void usage(const char *name)
 {
-	fprintf(stderr, _("Usage: %s [-hv] MIME-DIR\n"), name);
+	g_fprintf(stderr, _("Usage: %s [-hvV] MIME-DIR\n"), name);
 }
 
 static void free_type(gpointer data)
@@ -176,7 +200,7 @@ static Type *get_type(const char *name, GError **error)
 			return type;
 	}
 
-	g_print("* Warning: Unknown media type in type '%s'\n", name);
+	g_warning("Unknown media type in type '%s'\n", name);
 
 	return type;
 }
@@ -364,12 +388,6 @@ static gboolean process_freedesktop_node(Type *type, xmlNode *field,
 		if (localName)
 			xmlFree(localName);
 	}
-	else
-	{
-		g_set_error(error, MIME_ERROR, 0,
-			_("Unknown freedesktop.org field '%s'"),
-			field->name);
-	}
 	
 	return !*error;
 }
@@ -498,16 +516,16 @@ static void load_source_file(const char *filename)
 
 	if (root->ns == NULL || xmlStrcmp(root->ns->href, FREE_NS) != 0)
 	{
-		g_print("* Wrong namespace on document element\n"
-			"*   in '%s'\n"
-			"*   (should be %s)\n", filename, FREE_NS);
+		g_warning("* Wrong namespace on document element\n"
+			  "*   in '%s'\n"
+			  "*   (should be %s)\n", filename, FREE_NS);
 		goto out;
 	}
 	
 	if (strcmp((char *)root->name, "mime-info") != 0)
 	{
-		g_print("* Root element <%s> is not <mime-info>\n"
-			"*   (in '%s')\n", root->name, filename);
+		g_warning("* Root element <%s> is not <mime-info>\n"
+			  "*   (in '%s')\n", root->name, filename);
 		goto out;
 	}
 
@@ -551,12 +569,12 @@ static void load_source_file(const char *filename)
 
 		if (error)
 		{
-			g_print("* Error in type '%s/%s'\n"
-				"*   (in %s):\n"
-				"*   %s.\n",
-				type ? type->media : _("unknown"),
-				type ? type->subtype : _("unknown"),
-				filename, error->message);
+			g_warning("* Error in type '%s/%s'\n"
+				  "*   (in %s):\n"
+				  "*   %s.\n",
+				  type ? type->media : _("unknown"),
+				  type ? type->subtype : _("unknown"),
+				  filename, error->message);
 			g_error_free(error);
 		}
 	}
@@ -660,11 +678,11 @@ static void write_out_glob(gpointer key, gpointer value, gpointer data)
 	for ( ; list; list = list->next) {
 		type = (Type *)list->data;
 		if (strchr(pattern, '\n'))
-			g_print("* Glob patterns can't contain literal newlines "
-				"(%s in type %s/%s)\n", pattern,
-				type->media, type->subtype);
+			g_warning("* Glob patterns can't contain literal newlines "
+				  "(%s in type %s/%s)\n", pattern,
+				  type->media, type->subtype);
 		else
-			fprintf(stream, "%s/%s:%s\n",
+			g_fprintf(stream, "%s/%s:%s\n",
 				type->media, type->subtype, pattern);
 	}
 }
@@ -681,8 +699,12 @@ static void atomic_update(const gchar *pathname)
 
 	new_name = g_strndup(pathname, len - 4);
 
+#ifdef _WIN32
+	/* we need to remove the old file first! */
+	remove(new_name);
+#endif
 	if (rename(pathname, new_name))
-		g_warning("Failed to rename %s as %s\n", pathname, new_name);
+		g_warning("Failed to rename %s as %s , errno: %d\n", pathname, new_name, errno);
 
 	g_free(new_name);
 }
@@ -695,7 +717,11 @@ static void write_out_type(gpointer key, gpointer value, gpointer data)
 	char *media, *filename;
 
 	media = g_strconcat(mime_dir, "/", type->media, NULL);
+#ifdef _WIN32
+	mkdir(media);
+#else
 	mkdir(media, 0755);
+#endif
 
 	filename = g_strconcat(media, "/", type->subtype, ".xml.new", NULL);
 	g_free(media);
@@ -1279,9 +1305,12 @@ static void write_magic_children(FILE *stream, GList *matches, int indent)
 		Match *match = (Match *) next->data;
 
 		if (indent)
-			fprintf(stream, "%d>%ld=", indent, match->range_start);
+			g_fprintf(stream,
+				  "%d>%ld=",
+				  indent,
+				  match->range_start);
 		else
-			fprintf(stream, ">%ld=", match->range_start);
+			g_fprintf(stream, ">%ld=", match->range_start);
 
 		write16(stream, match->data_length);
 		fwrite(match->data, match->data_length, 1, stream);
@@ -1291,9 +1320,9 @@ static void write_magic_children(FILE *stream, GList *matches, int indent)
 			fwrite(match->mask, match->data_length, 1, stream);
 		}
 		if (match->word_size != 1)
-			fprintf(stream, "~%d", match->word_size);
+			g_fprintf(stream, "~%d", match->word_size);
 		if (match->range_length != 1)
-			fprintf(stream, "+%d", match->range_length);
+			g_fprintf(stream, "+%d", match->range_length);
 
 		fputc('\n', stream);
 
@@ -1304,7 +1333,7 @@ static void write_magic_children(FILE *stream, GList *matches, int indent)
 /* Write a whole Magic element to the 'magic' file */
 static void write_magic(FILE *stream, Magic *magic)
 {
-	fprintf(stream, "[%d:%s/%s]\n", magic->priority,
+	g_fprintf(stream, "[%d:%s/%s]\n", magic->priority,
 		magic->type->media, magic->type->subtype);
 
 	write_magic_children(stream, magic->matches, 0);
@@ -1346,7 +1375,7 @@ static void delete_old_types(const gchar *mime_dir)
 				path = g_strconcat(mime_dir, "/",
 						type_name, ".xml", NULL);
 #if 0
-				g_print("* Removing old info for type %s\n",
+				g_warning("* Removing old info for type %s\n",
 						path);
 #endif
 				unlink(path);
@@ -1469,15 +1498,15 @@ static void check_in_path_xdg_data(const char *mime_path)
 
 	if (stat(path, &path_info))
 	{
-		g_printerr("Can't stat '%s' directory: %s",
+		g_warning("Can't stat '%s' directory: %s",
 				path, g_strerror(errno));
 		goto out;
 	}
 
 	env = getenv("XDG_DATA_DIRS");
 	if (!env)
-		env = "/usr/local/share/:/usr/share/";
-	dirs = g_strsplit(env, ":", 0);
+		env = "/usr/local/share/"PATH_SEPARATOR"/usr/share/";
+	dirs = g_strsplit(env, PATH_SEPARATOR, 0);
 	g_return_if_fail(dirs != NULL);
 	for (n = 0; dirs[n]; n++)
 		;
@@ -1499,15 +1528,15 @@ static void check_in_path_xdg_data(const char *mime_path)
 
 	if (i == n)
 	{
-		g_print(_("\nNote that '%s' is not in the search path\n"
+		g_warning(_("\nNote that '%s' is not in the search path\n"
 			"set by the XDG_DATA_HOME and XDG_DATA_DIRS\n"
 			"environment variables, so applications may not\n"
 			"be able to find it until you set them. The\n"
 			"directories currently searched are:\n\n"), path);
-		g_print("- %s\n", dirs[n - 1]);
+		g_printerr("- %s\n", dirs[n - 1]);
 		for (i = 0; i < n - 1; i++)
-			g_print("- %s\n", dirs[i]);
-		g_print("\n");
+			g_printerr("- %s\n", dirs[i]);
+		g_printerr("\n");
 	}
 
 	for (i = 0; i < n; i++)
@@ -1632,7 +1661,7 @@ write_map_entry (gpointer key,
       offset = GPOINTER_TO_UINT (g_hash_table_lookup (map_data->pool, values[i]));
       if (offset == 0)
 	{
-	  g_printerr ("Missing string: '%s'\n", values[i]);
+	  g_warning ("Missing string: '%s'\n", values[i]);
 	  map_data->error = TRUE;  
 	}
       
@@ -1760,7 +1789,7 @@ write_parent_entry (gpointer key,
   offset = GPOINTER_TO_UINT (g_hash_table_lookup (map_data->pool, mimetype));
   if (offset == 0)
     {
-      g_printerr ("Missing string: '%s'\n", (gchar *)key);
+      g_warning ("Missing string: '%s'\n", (gchar *)key);
       map_data->error = TRUE;  
     }
 
@@ -1793,7 +1822,7 @@ write_parent_list (gpointer key,
       offset = GPOINTER_TO_UINT (g_hash_table_lookup (map_data->pool, parent));
       if (offset == 0)
 	{
-	  g_printerr ("Missing string: '%s'\n", parent);
+	  g_warning ("Missing string: '%s'\n", parent);
 	  map_data->error = TRUE;  
 	}
       
@@ -2014,7 +2043,7 @@ build_suffixes (gpointer key,
       
       if (suffix == NULL)
 	{
-	  g_printerr ("Glob '%s' is not valid UTF-8\n", glob);
+	  g_warning ("Glob '%s' is not valid UTF-8\n", glob);
 	  return;
 	}
 
@@ -2078,7 +2107,7 @@ write_suffix_entries (FILE        *cache,
       offset = GPOINTER_TO_UINT(g_hash_table_lookup (strings, entry->mimetype));
       if (offset == 0)
 	{
-	  g_printerr ("Missing string: '%s'\n", entry->mimetype);
+	  g_warning ("Missing string: '%s'\n", entry->mimetype);
 	  return FALSE;
 	}
     }
@@ -2172,7 +2201,7 @@ write_match (gpointer key,
   offset = GPOINTER_TO_UINT (g_hash_table_lookup (mdata->strings, mimetype));
   if (offset == 0)
     {
-      g_printerr ("Missing string: '%s'\n", mimetype);
+      g_warning ("Missing string: '%s'\n", mimetype);
       g_free (mimetype);
       mdata->error = TRUE;
       return;
@@ -2556,7 +2585,7 @@ write_cache (FILE *cache)
   offset = 0;
   if (!write_header (cache, 0, 0, 0, 0, 0, 0, 0, &offset))
     {
-      g_printerr ("Failed to write header\n");
+      g_warning ("Failed to write header\n");
       return FALSE;
     }
 
@@ -2566,67 +2595,67 @@ write_cache (FILE *cache)
 
   if (!write_strings (cache, strings, &offset))
     {
-      g_printerr ("Failed to write strings\n");
+      g_warning ("Failed to write strings\n");
       return FALSE;
     }
-  g_print ("Wrote %d strings at %x - %x\n", 
+  g_message ("Wrote %d strings at %x - %x\n", 
 	   g_hash_table_size (strings), strings_offset, offset);
 
   alias_offset = offset;
   if (!write_alias_cache (cache, strings, &offset))
     {
-      g_printerr ("Failed to write alias list\n");
+      g_warning ("Failed to write alias list\n");
       return FALSE;
     }
-  g_print ("Wrote aliases at %x - %x\n", alias_offset, offset);
+  g_message ("Wrote aliases at %x - %x\n", alias_offset, offset);
 
   parent_offset = offset;
   if (!write_parent_cache (cache, strings, &offset))
     {
-      g_printerr ("Failed to write parent list\n");
+      g_warning ("Failed to write parent list\n");
       return FALSE;
     }
-  g_print ("Wrote parents at %x - %x\n", parent_offset, offset);
+  g_message ("Wrote parents at %x - %x\n", parent_offset, offset);
 
   literal_offset = offset;
   if (!write_literal_cache (cache, strings, &offset))
     {
-      g_printerr ("Failed to write literal list\n");
+      g_warning ("Failed to write literal list\n");
       return FALSE;
     }
-  g_print ("Wrote literal globs at %x - %x\n", literal_offset, offset);
+  g_message ("Wrote literal globs at %x - %x\n", literal_offset, offset);
 
   suffix_offset = offset;
   if (!write_suffix_cache (cache, strings, &offset))
     {
-      g_printerr ("Failed to write suffix list\n");
+      g_warning ("Failed to write suffix list\n");
       return FALSE;
     }
-  g_print ("Wrote suffix globs at %x - %x\n", suffix_offset, offset);
+  g_message ("Wrote suffix globs at %x - %x\n", suffix_offset, offset);
 
   glob_offset = offset;
   if (!write_glob_cache (cache, strings, &offset))
     {
-      g_printerr ("Failed to write glob list\n");
+      g_warning ("Failed to write glob list\n");
       return FALSE;
     }
-  g_print ("Wrote full globs at %x - %x\n", glob_offset, offset);
+  g_message ("Wrote full globs at %x - %x\n", glob_offset, offset);
 
   magic_offset = offset;
   if (!write_magic_cache (cache, strings, &offset))
     {
-      g_printerr ("Failed to write magic list\n");
+      g_warning ("Failed to write magic list\n");
       return FALSE;
     }
-  g_print ("Wrote magic at %x - %x\n", magic_offset, offset);
+  g_message ("Wrote magic at %x - %x\n", magic_offset, offset);
 
   namespace_offset = offset;
   if (!write_namespace_cache (cache, strings, &offset))
     {
-      g_printerr ("Failed to write namespace list\n");
+      g_warning ("Failed to write namespace list\n");
       return FALSE;
     }
-  g_print ("Wrote namespace list at %x - %x\n", namespace_offset, offset);
+  g_debug ("Wrote namespace list at %x - %x\n", namespace_offset, offset);
 
   rewind (cache);
   offset = 0; 
@@ -2636,7 +2665,7 @@ write_cache (FILE *cache)
 		     suffix_offset, glob_offset, magic_offset, 
 		     namespace_offset, &offset))
     {
-      g_printerr ("Failed to rewrite header\n");
+      g_warning ("Failed to rewrite header\n");
       return FALSE;
     }
 
@@ -2646,13 +2675,30 @@ write_cache (FILE *cache)
 }
 
 
+static FILE *
+open_or_die(const char *filename)
+{
+	FILE *stream = fopen(filename, "wb");
+
+	if (!stream)
+	{
+		g_printerr("Failed to open '%s' for writing\n", filename);
+		exit(EXIT_FAILURE);
+	}
+
+	return stream;
+}
+
 int main(int argc, char **argv)
 {
 	char *mime_dir = NULL;
 	char *package_dir = NULL;
 	int opt;
 
-	while ((opt = getopt(argc, argv, "hv")) != -1)
+	/* Install the filtering log handler */
+	g_log_set_default_handler(g_log_handler, NULL);
+
+	while ((opt = getopt(argc, argv, "hvV")) != -1)
 	{
 		switch (opt)
 		{
@@ -2663,12 +2709,16 @@ int main(int argc, char **argv)
 				usage(argv[0]);
 				return EXIT_SUCCESS;
 			case 'v':
-				fprintf(stderr,
-					"update-mime-database (" PACKAGE ") "
-					VERSION "\n" COPYING);
+				g_fprintf(stderr,
+					  "update-mime-database (" PACKAGE ") "
+					  VERSION "\n" COPYING);
 				return EXIT_SUCCESS;
+			case 'V':
+				enabled_log_levels |= G_LOG_LEVEL_MESSAGE
+						      | G_LOG_LEVEL_INFO;
+				break;
 			default:
-				abort();
+				return EXIT_FAILURE;
 		}
 	}
 
@@ -2696,16 +2746,16 @@ int main(int argc, char **argv)
 
 	if (access(mime_dir, W_OK))
 	{
-		g_printerr(_("%s: I don't have write permission on %s.\n"
+		g_warning(_("%s: I don't have write permission on %s.\n"
 			     "Try rerunning me as root.\n"), argv[0], mime_dir);
 		return EXIT_FAILURE;
 	}
 
-	g_print("***\n* Updating MIME database in %s...\n", mime_dir);
+	g_message("Updating MIME database in %s...\n", mime_dir);
 
 	if (access(package_dir, F_OK))
 	{
-		fprintf(stderr,
+		g_fprintf(stderr,
 			_("Directory '%s' does not exist!\n"), package_dir);
 		return EXIT_FAILURE;
 	}
@@ -2733,13 +2783,10 @@ int main(int argc, char **argv)
 		FILE *globs;
 		char *globs_path;
 		globs_path = g_strconcat(mime_dir, "/globs.new", NULL);
-		globs = fopen(globs_path, "wb");
-		if (!globs)
-			g_error("Failed to open '%s' for writing\n",
-				globs_path);
-		fprintf(globs,
-			"# This file was automatically generated by the\n"
-			"# update-mime-database command. DO NOT EDIT!\n");
+		globs = open_or_die(globs_path);
+		g_fprintf(globs,
+			  "# This file was automatically generated by the\n"
+			  "# update-mime-database command. DO NOT EDIT!\n");
 		g_hash_table_foreach(globs_hash, write_out_glob, globs);
 		fclose(globs);
 
@@ -2752,10 +2799,7 @@ int main(int argc, char **argv)
 		char *magic_path;
 		int i;
 		magic_path = g_strconcat(mime_dir, "/magic.new", NULL);
-		stream = fopen(magic_path, "wb");
-		if (!stream)
-			g_error("Failed to open '%s' for writing\n",
-					magic_path);
+		stream = open_or_die(magic_path);
 		fwrite("MIME-Magic\0\n", 1, 12, stream);
 
 		if (magic_array->len)
@@ -2777,12 +2821,9 @@ int main(int argc, char **argv)
 		char *ns_path;
 
 		ns_path = g_strconcat(mime_dir, "/XMLnamespaces.new", NULL);
-		stream = fopen(ns_path, "wb");
-		if (!stream)
-			g_error("Failed to open '%s' for writing\n",
-					ns_path);
-
+		stream = open_or_die(ns_path);
 		write_namespaces(stream);
+		fclose(stream);
 
 		atomic_update(ns_path);
 		g_free(ns_path);
@@ -2793,12 +2834,9 @@ int main(int argc, char **argv)
 		char *path;
 		
 		path = g_strconcat(mime_dir, "/subclasses.new", NULL);
-		stream = fopen(path, "wb");
-		if (!stream)
-			g_error("Failed to open '%s' for writing\n",
-				path);
-		
+		stream = open_or_die(path);
 		write_subclasses(stream);
+		fclose(stream);
 
 		atomic_update(path);
 		g_free(path);
@@ -2809,12 +2847,9 @@ int main(int argc, char **argv)
 		char *path;
 		
 		path = g_strconcat(mime_dir, "/aliases.new", NULL);
-		stream = fopen(path, "wb");
-		if (!stream)
-			g_error("Failed to open '%s' for writing\n",
-				path);
-		
+		stream = open_or_die(path);
 		write_aliases(stream);
+		fclose(stream);
 
 		atomic_update(path);
 		g_free(path);
@@ -2825,12 +2860,9 @@ int main(int argc, char **argv)
 		char *path;
 		
 		path = g_strconcat(mime_dir, "/mime.cache.new", NULL);
-		stream = fopen(path, "wb");
-		if (!stream)
-			g_error("Failed to open '%s' for writing\n",
-				path);
-		
+		stream = open_or_die(path);
 		write_cache(stream);
+		fclose(stream);
 
 		atomic_update(path);
 		g_free(path);
@@ -2845,8 +2877,6 @@ int main(int argc, char **argv)
 	g_hash_table_destroy(subclass_hash);
 	g_hash_table_destroy(alias_hash);
 
-	g_print("***\n");
-
 	check_in_path_xdg_data(mime_dir);
 
 	return EXIT_SUCCESS;
